// Copyright (C) 2003  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.

#ifdef DLIB_ALL_SOURCE_END
#include "dlib_basic_cpp_build_tutorial.txt"
#endif

#ifndef DLIB_THREADs_
#define DLIB_THREADs_

#include "threads/threads_kernel.h"

#include "threads/auto_mutex_extension.h"
#include "threads/auto_unlock_extension.h"
#include "threads/create_new_thread_extension.h"
#include "threads/multithreaded_object_extension.h"
#include "threads/rmutex_extension.h"
#include "threads/rsignaler_extension.h"
#include "threads/threaded_object_extension.h"
#include "threads/thread_specific_data_extension.h"
#include "threads/thread_function_extension.h"
#include "threads/thread_pool_extension.h"
#include "threads/read_write_mutex_extension.h"
#include "threads/parallel_for_extension.h"
#include "threads/async.h"

#endif // DLIB_THREADs_

